# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Union
from pathlib import Path
import json
import yaml


def check_path(
    p: Union[Path, str],
    is_dir: bool = False,
    make_parents: bool = True,
    make_dir_if_dir: bool = True,
) -> Path:
    """Check a user supplied path (str or Path), ensuring parents exist etc

    Parameters
    ----------
    p : Union[Path, str]
        File or directory path supplied by user
    is_dir : bool, optional
        If the path supplied by the user should be a directory, then set it as such by assigning is_dir to true, by default False
    make_parents : bool, optional
        If the parent directories of the supplied path do not exist, the make them, by default True
    make_dir_if_dir : bool, optional
        If the supplied path is a directory, but it does not exist, the make it, by default True

    Returns
    -------
    Path
        Path object pointing to the user supplied path, with parents made (if requested), and the directory itself made (if a directory and requested)

    Raises
    ------
    ValueError
        Parent does not exist, and make_parents was False
    ValueError
        Passed path was not a string or pathlib.Path
    """
    if isinstance(p, str):
        p = Path(p)
    if isinstance(p, Path):
        p = p.resolve()

        # Is dir
        if is_dir:
            if p.exists():
                return p
            if make_parents and not p.parent.exists():
                p.parent.mkdir(parents=True)
            if make_dir_if_dir and not p.exists():
                p.mkdir(parents=True)
            return p
        # Is file
        else:
            if not p.parent.exists():
                if make_parents:
                    p.mkdir(parents=True)
                    return p
                else:
                    raise ValueError(
                        f"{p.parent} does not exist, and make_parents was False when specifying the file {p}"
                    )
            else:
                return p
    else:
        raise ValueError(
            f"File/directory ({p}) passed as argument was not a string or Path, it was {type(p)}"
        )


def load_dict(file_path: Union[str, Path, None], cast_none_to_dict=False):
    if cast_none_to_dict and file_path is None:
        return {}
    if isinstance(file_path, dict):
        return file_path
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if not isinstance(file_path, (str, Path)):
        raise TypeError(
            f"Argument 'file_path' to load_dict function was not None, str, or Path. It was {type(file_path)}"
        )
    if not file_path.exists():
        raise FileNotFoundError(f"The file '{file_path}' does not exist")
    if file_path.suffix in [".yml", ".yaml"]:
        return yaml.safe_load(open(file_path))
    elif file_path.suffix == ".json":
        return json.load(open(file_path))
    else:
        raise ValueError(
            f"File had invalid extension ('{file_path.suffix}'), please supply a .yml, .yaml, or .json file"
        )



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

import shutil
import pandas as pd
import numpy as np
from phenonaut.output.heatmap import write_heatmap_from_df
from phenonaut.phenonaut import Phenonaut
from pathlib import Path
import tempfile


def test_get_y_from_iris_dataset(dataset_iris):
    from phenonaut.predict.predict_utils import get_y_from_target

    assert get_y_from_target(dataset_iris).shape == (150,)


def test_write_heatmap():
    with tempfile.TemporaryDirectory() as tmp_dir:
        df = pd.DataFrame(
            data=np.array([[0.1, 0.2, 0.3], [0.1, 0.2, 0.3], [0.5, 0.6, 0.7]]),
            columns=["X1", "X2", "X3"],
            index=["Y1", "Y2", "Y3"],
        )
        write_heatmap_from_df(df, "Test heatmap", Path(tmp_dir) / "tmp_heatmap.png")
        assert (Path(tmp_dir) / "tmp_heatmap.png").stat().st_size > 10000


def test_instantiate_dave():
    from phenonaut.predict.default_predictors.pytorch_models.dave import _DAVE_model

    d = _DAVE_model(view_sizes=(100, 150), embedding_size=32, n_hidden=2)
    assert d.topology is not None


def test_predict_iris_class_1_view(phenonaut_object_iris_2_views, dataset_iris):
    dirpath = tempfile.mkdtemp()
    print(dirpath)
    phe = Phenonaut(dataset_iris)
    import phenonaut.predict

    phenonaut.predict.profile(phe, dirpath)
    shutil.rmtree(dirpath)


def test_predict_iris_class_2_views(phenonaut_object_iris_2_views):
    dirpath = tempfile.mkdtemp()
    print(dirpath)
    phe = phenonaut_object_iris_2_views
    import phenonaut.predict

    phenonaut.predict.profile(phe, dirpath)
    shutil.rmtree(dirpath)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from copy import deepcopy
from matplotlib import pyplot as plt

from pathlib import Path
from typing import Tuple, Union, Optional, List
import numpy as np
from pathlib import Path
from typing import Union, Optional
import numpy as np
import pandas as pd
import seaborn as sns

from matplotlib.ticker import StrMethodFormatter
from matplotlib.colors import Colormap


def write_heatmap_from_df(
    df: pd.DataFrame,
    title: str,
    output_file: Union[str, Path],
    axis_labels: tuple[str, str] = ("", ""),
    annot: Union[bool, np.ndarray] = True,
    transpose: Union[bool, None] = None,
    standard_deviations_df: Optional[pd.DataFrame] = None,
    figsize: Optional[Union[Tuple, List]] = None,
    figure_dpi: int = 300,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    annotation_fontsize: int = 10,
    annotation_best_fontsize: int = 12,
    lower_is_better: bool = False,
    annotation_format_string_value=".2f",
    annotation_format_string_std=".3f",
    highlight_best: bool = True,
    pallet_name: str = "seagreen",
    put_cbar_on_left: bool = False,
    sns_colour_palette: Optional[Colormap] = None,
) -> None:
    """Write heatmap from pd.DataFrame

    Write a PNG or SVG heatmap image to a file uisng seaborn to read from
    a pd.DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame containing values for the heatmap
    title : str
        Heatmap title
    output_file : Union[str, Path]
        Output path for the image, must have the extension .png or .svg.
    axis_labels : tuple[str, str], optional
        Tuple denoting labels to use for each axis, by default ("", "")
    annot : Union[bool, np.ndarray], optional
        If True, the the values of each heatmap cell/box are writen alongside
        applying the correct background colouring to the cell/box.
        By default True.
    transpose : Union[bool, None], optional
        Transpose the heatmap, by default None
    standard_deviations_df : Optional[pd.DataFrame], optional
        Optionally, supply a second pd.DataFrame containing standard deviations
        which may be included alongside the annotations within each cell/box.
        By default None.
    figsize : Optional[Union[Tuple, List]], optional
        Optional argument for setting the figure size. If None, then a target
        output width is set at 10, and the height calculated through examination
        of the number of elements, keeping square aspect ratios for each
        cell/box. By default None.
    figure_dpi : int, optional
        Resolution (dots per inch) at which to save the image (only useful if
        writing PNGs), by default 300.
    vmin : Optional[float], optional
        Minimum value to use for defining the colour/shading range. If None,
        then it is set to the lowest value present in the input DataFrame. By
        default None.
    vmax : Optional[float], optional
        Maximum value to use for defining the colour/shading range. If None,
        then it is set to the highest value present in the input DataFrame. By
        default None.
    annotation_fontsize : int, optional
        Font size for annotations, by default 10
    annotation_best_fontsize : int, optional
        Font size for the annotation used to highlight the best value present
        within the input DataFrame, by default 12
    lower_is_better : bool, optional
        Used for selecting the best value within the dataframe. If True, then
        the best heatmap value is taken to be the lowest. By default False.
    annotation_format_string_value : str, optional
        Optional formatting string for annotation f-strings, by default ".2f".
    annotation_format_string_std : str, optional
        Optional formatting string for standard deviation annotation f-strings,
        by default ".3f".
    highlight_best : bool, optional
        If True, then highlight the best result, with best being determined
        through use of the 'lower_is_better' keyword. By default True.
    pallet_name : str, optional
        Pallet to use for the colour bar. By default "seagreen".
    put_cbar_on_left : bool, optional
        If True, then position the colour bar on the left. If False, then
        position the colour bar on the right. By default False.
    sns_colour_palette : Optional[Colormap], optional
        Optionally, supply a seaborn colour map for use in the colour bar. By
        default None.
    """

    padding = {
        "top": 0.25,
        "bottom": 1.24,
        "element_height": 0.60,
    }
    if figsize is None:
        figsize = (
            10.0,
            padding["top"] + padding["bottom"] + padding["element_height"] * df.shape[0],
        )
    if (transpose is None and df.shape[0] > df.shape[1]) or transpose:
        df = deepcopy(df.transpose())
    # Turn column and row labels with spaces into multiline items
    df = df.rename(
        columns={c: c.replace(" ", "\n") for c in df.columns if isinstance(c, str)},
        index={
            ind: ind.replace(" ", "\n").replace(":", "\n")
            for ind in df.index
            if isinstance(ind, str)
        },
    )

    if not isinstance(annot, np.ndarray):
        if annot and standard_deviations_df is not None:
            annot = (
                np.array(
                    [
                        f"{val:{annotation_format_string_value}}\n({std:{annotation_format_string_std}})"
                        for val, std in zip(
                            df.values.ravel(), standard_deviations_df.values.ravel()
                        )
                    ]
                )
                .reshape(df.values.shape)
                .tolist()
            )
        else:
            annot = (
                np.array([f"{val:{annotation_format_string_value}}" for val in df.values.ravel()])
                .reshape(df.values.shape)
                .tolist()
            )

    fig, ax = plt.subplots(figsize=figsize, facecolor="w")
    if highlight_best:
        mask = df.values == np.nanmax(df.values)
        if lower_is_better:
            mask = ~mask
        negated_mask = ~mask
        sns.heatmap(
            df,
            cmap=sns.light_palette(pallet_name, as_cmap=True, reverse=lower_is_better)
            if sns_colour_palette is None
            else sns_colour_palette,
            mask=mask,
            annot=annot,
            linewidths=0.02,
            ax=ax,
            fmt="",
            annot_kws={"fontsize": annotation_fontsize},
            vmin=np.nanmin(df.values) if vmin is None else vmin,
            vmax=np.nanmax(df.values) if vmax is None else vmax,
            cbar_kws=dict(location="left") if put_cbar_on_left else None,
        )
        sns.heatmap(
            df,
            cmap=sns.light_palette(pallet_name, as_cmap=True, reverse=lower_is_better)
            if sns_colour_palette is None
            else sns_colour_palette,
            mask=negated_mask,
            annot=annot,
            linewidths=0.02,
            ax=ax,
            fmt="",
            annot_kws={"fontsize": annotation_best_fontsize, "style": "italic", "weight": "bold"},
            cbar=False,
            vmin=np.nanmin(df.values) if vmin is None else vmin,
            vmax=np.nanmax(df.values) if vmax is None else vmax,
            cbar_kws=dict(location="left") if put_cbar_on_left else None,
        )
    else:
        sns.heatmap(
            df,
            cmap=sns.light_palette(pallet_name, as_cmap=True, reverse=lower_is_better)
            if sns_colour_palette is None
            else sns_colour_palette,
            annot=annot,
            linewidths=0.02,
            ax=ax,
            fmt="",
            annot_kws={"fontsize": annotation_fontsize},
            vmin=np.nanmin(df.values) if vmin is None else vmin,
            vmax=np.nanmax(df.values) if vmax is None else vmax,
            cbar_kws=dict(location="left") if put_cbar_on_left else None,
        )

    plt.xlabel(axis_labels[0])
    plt.ylabel(axis_labels[1])
    ax.set_title(title)
    plt.yticks(rotation=15)
    fig = ax.get_figure()
    plt.tight_layout()
    fig.savefig(output_file, dpi=figure_dpi)
    plt.close()



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from collections import namedtuple
from pathlib import Path
from typing import List, Optional, Union
from ..data import Dataset
import pandas as pd
from .base import PackagedDataset
import gzip
import pandas as pd
import numpy as np
import h5py
import datetime
import json


class CMAP(PackagedDataset):
    """CMAP (Level 5)- ConnectivityMap dataset - https://clue.io/

    CMAP is a repository of L1000 profiles measured from small molecule and
    crisper perturbations. This CMAP packaged dataset supplies an interface
    to querying this data and allows access to the data through a Phenonaut
    dataset object. This PackagedDataset is for level 5 data (merged
    profiles).

    Data supplied by CMAP is in their own GCTX format files which are HDF5
    files with data residing in specific paths. Rather than have Phenonaut
    rely on their supplied library, we simply read the file with standard
    HDF5 tools. Additionally, we merge sig_info data, assigning
    perturbation information, bringing in the following columns:
    'pert_id', 'pert_iname', 'pert_type', 'cell_id', 'pert_idose',
    'pert_itime', and 'distil_id'.

    in the pert_type column, and taking the following values:

    ctl_vehicle : DMSO

    trt_cp      : compound treatment

    trt_xpr     : crisper treatment

    If the pert_type is trt_cp, then pert_idose gives the compound
    concentration.  In the supplied cmap data, the field is a string
    containing for example, "3.33 um". This dataloader changes this to a
    float field without the µM (um as written) prefix - units are µM. If
    the pert_type is ctl_vehicle or trt_xpr, then CMAP supplied data with
    -666 in the pert_idose field. This is changed to np.nan.

    Further field information can be found here:
    https://clue.io/connectopedia/glossary

    This PackagedDataset provides supplies the following pd.DataFrames
    (queryable by calling the inherited ".keys" method):

    * creation_date
        the date on which the h5 file was written.
    * df
        the main dataframe containing L1000 data, merged with sig_info to give a more complete view of the dataset.
    * gene_info
        Information on the genes - allows translation of L1000 gene number in df to gene name and more.
    * pert_info
        Perturbation info.  Using the df, which contains the merged sig_info data, compound ids can be used to query this dataframe and molecule smiles etc returned.
    * sig_metrics
        Additional metrics on the profiles recorded in df.
    * inst_info
        Information on plate barcodes etc from which profiles derived.

    Parameters
    ----------
    root : Union[Path, str]
        Local directory containing the prepared dataset. If the dataset is
        not found here and the argument download=True is not given, then an
        error is raised. If download=True and the processed dataset is
        absent, then it is downloaded the directory pointed at by the
        'raw_data' argument detailed below. If raw_data_dir is a
        non-absolute path, such as a single directory, then it is created as
        a subdirectory of this root directory.
    download : bool, optional
        If true and the processed dataset is not found in the root
        directory, then the dataset is downloaded and processed.
        By default False.
    raw_data_dir : Optional[Union[Path, str]], optional
        If downloading and preparing the dataset, then a directory for the
        raw data may be specified. If a non-absolute location is given,
        then it is created in a subdirectory of the root directory specified
        as the first argument. Absolute paths may be used to place raw
        datafiles and intermediates in another location, such as scratch
        disks etc.  Only has an effect if downloading and rebuilding the
        PackagedDataset.by default Path("raw_data").
    landmark_only : bool
        If True, then only return landmark genes, essentially removing all
        inferred gene abundances. This is likely the most useful for the
        majority of tasks ashighly correlated abundances simply adds to
        colinearity.  Only has an effect if rebuilding the PackagedDataset.
        By default, True.
    """

    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
        landmark_only: bool = True,
    ):

        super().__init__(root, raw_data_dir)
        self.name = "CMAP"
        self.landmark_only = landmark_only
        self.processed_h5_file = self.root / "cmap_phenonaut.h5"

        if not download and not self.processed_h5_file.exists():
            raise FileNotFoundError(
                f"CMAP PackagedDataset was instructed not to download the raw data, yet the given preprocessed Phenonaut file ({self.processed_h5_file}) does not exist"
            )

        # If the dataset is missing, get it
        self._call_if_file_missing(
            self.processed_h5_file, self._make_merged_data_and_sig_info, None
        )

        store = pd.HDFStore(self.processed_h5_file)
        required_keys = [
            "/creation_date",
            "/df",
            "/gene_info",
            "/pert_info",
            "/sig_metrics",
            "/inst_info",
        ]
        if not all(rk in store.keys() for rk in required_keys):
            raise KeyError(
                f"{self.processed_h5_file.name} does not contain the correct df keys, needed {required_keys}, but got {store.keys()}"
            )

        for k in [key for key in required_keys if not key == "/df"]:
            self.register_df_key(k[1:])
        self.register_df_key("landmark_gene_ids")

        self.register_ds_key("ds")

        store.close()

    def _make_merged_data_and_sig_info(self):
        """Make main CMAP HDF5 file

        Downloads the large compressed data file, decompresses it, extracts the
        data into a pd.DataFrame, merges the sig_info data file which includes
        data on the perturbation type.
        """

        # Make sure the required files to make the dataset exist
        DownloadableFileInfo = namedtuple(
            "DownloadableFileInfo", ["filename", "compressed_filename", "url"]
        )
        required_files = [
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
            ),
        ]
        for req_file in required_files:
            self._call_if_file_missing(
                self.root / req_file.filename,
                self._download,
                {
                    "source": req_file.url,
                    "destination": self.root / req_file.compressed_filename,
                    "extract": True,
                },
            )

        # Make sure the large GCTX file is present
        CMAP_GCTX_GZ_URL = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_Level5_COMPZ_n118050x12328_2017-03-06.gctx.gz"
        CMAP_GCTX_GZ_FILENAME = "GSE70138_Broad_LINCS_Level5_COMPZ_n118050x12328_2017-03-06.gctx"
        self._call_if_file_missing(
            self.root / CMAP_GCTX_GZ_FILENAME,
            self._download,
            {
                "source": CMAP_GCTX_GZ_URL,
                "destination": self.root / (CMAP_GCTX_GZ_FILENAME + ".gz"),
                "extract": True,
            },
        )

        # Load in CMAP large dataset from GCTX file, reading it as an HDF5 file, and then walking the correct path.
        f = h5py.File(self.root / CMAP_GCTX_GZ_FILENAME)
        print(f"{datetime.datetime.now()} Extracting large matrix... ")
        self._df = pd.DataFrame(
            data=f["0"]["DATA"]["0"]["matrix"],
            columns=f["0"]["META"]["ROW"]["id"],
            index=f["0"]["META"]["COL"]["id"],
        )
        self._df.index = self._df.index.map(lambda x: x.decode("utf8"))
        self._df.columns = self._df.columns.map(lambda x: x.decode("utf8"))
        print(f"{datetime.datetime.now()} ..done")

        # Here, we generate cmap_phenonaut.h5, containing a merged df
        print(f"{datetime.datetime.now()} Merging")
        self._sig_info = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        pert_dose = (
            self._sig_info["pert_idose"].apply(lambda x: x.replace(" um", "").strip()).astype(float)
        )
        self._sig_info["pert_idose"] = np.where(pert_dose < 0, np.nan, pert_dose)
        self._df = self._df.merge(self._sig_info, left_index=True, right_index=True, copy=False)

        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=0)
        h5_store["df"] = self._df

        h5_store["gene_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        h5_store["sig_metrics"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )
        h5_store["pert_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )

        h5_store["inst_info"] = pd.read_csv(
            self.root / "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt",
            sep="\t",
            index_col=[0],
        )

        h5_store["creation_date"] = pd.Series(str(datetime.datetime.now()))
        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        if key == "landmark_gene_ids":
            store = pd.HDFStore(self.processed_h5_file)
            df = store["/gene_info"]
            return df.query("pr_is_lm == 1").index.to_numpy(dtype=str)

        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        if key == "ds":
            key = "df"
        store = pd.HDFStore(self.processed_h5_file)
        pert_info_columns = [
            "pert_id",
            "pert_iname",
            "pert_type",
            "cell_id",
            "pert_idose",
            "pert_itime",
            "distil_id",
        ]
        if self.landmark_only:
            columns_needed = list(pert_info_columns)
            landmark_gene_list = self.get_df("landmark_gene_ids")
            columns_needed.extend(landmark_gene_list)
            df = store["/" + key][columns_needed]
        else:
            df = store["/" + key]
        features = [f for f in list(df.columns) if f not in pert_info_columns]
        ds = Dataset("cmap", df, {"features": features})
        store.close()
        return ds


class CMAP_Level4(PackagedDataset):
    """CMAP (Level4) - ConnectivityMap dataset - https://clue.io/

    CMAP is a repository of L1000 profiles measured from small molecule and
    crisper perturbations. This CMAP packaged dataset supplies an interface
    to querying this data and allows access to the data through a Phenonaut
    dataset object.

    Data supplied by CMAP is in their own GCTX format files which are HDF5
    files with data residing in specific paths. Rather than have Phenonaut
    rely on their supplied library, we simply read the file with standard
    HDF5 tools. Additionally, we merge sig_info data, assigning
    perturbation information, bringing in the following columns:
    'pert_id', 'pert_iname', 'pert_type', 'cell_id', 'pert_idose',
    'pert_itime', and 'distil_id'.

    in the pert_type column, and taking the following values:

    ctl_vehicle : DMSO

    trt_cp      : compound treatment

    trt_xpr     : crisper treatment

    If the pert_type is trt_cp, then pert_idose gives the compound
    concentration.  In the supplied cmap data, the field is a string
    containing for example, "3.33 um". This dataloader changes this to a
    float field without the µM (um as written) prefix - units are µM. If
    the pert_type is ctl_vehicle or trt_xpr, then CMAP supplied data with
    -666 in the pert_idose field. This is changed to np.nan.

    Further field information can be found here:
    https://clue.io/connectopedia/glossary

    This PackagedDataset provides supplies the following pd.DataFrames
    (queryable by calling the inherited ".keys" method):

    * creation_date
        the date on which the h5 file was written.
    * df
        the main dataframe containing L1000 data, merged with sig_info to give a more complete view of the dataset.
    * gene_info
        Information on the genes - allows translation of L1000 gene number in df to gene name and more.
    * pert_info
        Perturbation info.  Using the df, which contains the merged sig_info data, compound ids can be used to query this dataframe and molecule smiles etc returned.
    * sig_metrics
        Additional metrics on the profiles recorded in df.
    * inst_info
        Information on plate barcodes etc from which profiles derived.

    Parameters
    ----------
    root : Union[Path, str]
        Local directory containing the prepared dataset. If the dataset is
        not found here and the argument download=True is not given, then an
        error is raised. If download=True and the processed dataset is
        absent, then it is downloaded the directory pointed at by the
        'raw_data' argument detailed below. If raw_data_dir is a
        non-absolute path, such as a single directory, then it is created as
        a subdirectory of this root directory.
    download : bool, optional
        If true and the processed dataset is not found in the root
        directory, then the dataset is downloaded and processed.
        By default False.
    raw_data_dir : Optional[Union[Path, str]], optional
        If downloading and preparing the dataset, then a directory for the
        raw data may be specified. If a non-absolute location is given,
        then it is created in a subdirectory of the root directory specified
        as the first argument. Absolute paths may be used to place raw
        datafiles and intermediates in another location, such as scratch
        disks etc. Only has an effect if downloading and rebuilding the
        PackagedDataset. By default Path("raw_data").
    landmark_only : bool
        If True, then only return landmark genes, essentially removing all
        inferred gene abundances. This is likely the most useful for the
        majority of tasks ashighly correlated abundances simply adds to
        colinearity. Only has an effect if rebuilding the PackagedDataset.
        By default, True.
    allowed_treatment_types: Union[str, list[str]]
        Often, only compound treatments are needed for analysis, and so we
        include only treatments with pert_type of trt_cp or ctl_vehicle to
        allow compounds and DMSO vehicle only. This can be expanded to
        include crispr treatments with the inclusion of "trt_xpr", see
        https://clue.io/connectopedia/perturbagen_types_and_controls for
        further information on treatment types and possible values.
        Only has an effect if rebuilding the PackagedDataset.
        By default ['trt_cp','ctl_vehicle']
    allowed_treatment_times: Union[str, list[str]]
        Often, we are only interested in examining compound treatments
        after 24hrs, as there is an abundance of these measurements in the
        CMAP database. 24 hr treatments commonly have the pert_itime value
        '24 h'. Only has an effect if rebuilding the PackagedDataset. By
        default '24 h'.


    """

    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
        landmark_only: bool = True,
        allowed_treatment_types: Union[str, list[str]] = ["trt_cp", "ctl_vehicle"],
        allowed_treatment_times: Union[str, list[str]] = "24 h",
    ):

        super().__init__(root, raw_data_dir)
        self.name = "CMAP_Level4"
        self.landmark_only = landmark_only
        self.processed_h5_file = self.root / "cmap_level4_phenonaut.h5"

        if isinstance(allowed_treatment_types, str):
            allowed_treatment_types = [allowed_treatment_types]
        self.allowed_treatment_types = allowed_treatment_types
        if isinstance(allowed_treatment_times, str):
            allowed_treatment_times = [allowed_treatment_times]
        self.allowed_treatment_times = allowed_treatment_times

        if not download and not self.processed_h5_file.exists():
            raise FileNotFoundError(
                f"CMAP PackagedDataset was instructed not to download the raw data, yet the given preprocessed Phenonaut file ({self.processed_h5_file}) does not exist"
            )

        # If the dataset is missing, get it
        self._call_if_file_missing(
            self.processed_h5_file, self._make_merged_data_and_sig_info, None
        )

        store = pd.HDFStore(self.processed_h5_file)
        required_keys = [
            "/creation_date",
            "/df",
            "/gene_info",
            "/pert_info",
            "/sig_metrics",
            "/inst_info",
            "/features",
            "/allowed_treatment_times",
            "/allowed_treatment_types",
        ]
        if not all(rk in store.keys() for rk in required_keys):
            raise KeyError(
                f"{self.processed_h5_file.name} does not contain the correct df keys, needed {required_keys}, but got {store.keys()}"
            )

        for k in [key for key in required_keys if not key == "/df"]:
            self.register_df_key(k[1:])
        self.register_ds_key("ds")
        store.close()

    def _make_merged_data_and_sig_info(self):
        """Make main CMAP HDF5 file

        Downloads the large compressed data file, decompresses it, extracts the
        data into a pd.DataFrame, merges the sig_info data file which includes
        data on the perturbation type.

        """

        CMAP_GCTX_GZ_FILENAME = (
            "GSE70138_Broad_LINCS_Level4_ZSPCINF_mlr12k_n345976x12328_2017-03-06.gctx.gz"
        )
        # Make sure the required files to make the dataset exist
        DownloadableFileInfo = namedtuple(
            "DownloadableFileInfo", ["filename", "compressed_filename", "url"]
        )
        required_files = [
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt",
                "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt",
                "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
                "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
            ),
            DownloadableFileInfo(
                CMAP_GCTX_GZ_FILENAME.replace(".gz", ""),
                CMAP_GCTX_GZ_FILENAME,
                f"https://ftp.ncbi.nlm.nih.gov/geo/series/GSE70nnn/GSE70138/suppl/{CMAP_GCTX_GZ_FILENAME}",
            ),
        ]
        for req_file in required_files:
            self._call_if_file_missing(
                self.raw_data_dir / req_file.filename,
                self._download,
                {
                    "source": req_file.url,
                    "destination": self.raw_data_dir / req_file.compressed_filename,
                    "extract": False,
                },
            )

        sig_info = pd.read_csv(
            self.raw_data_dir / "GSE70138_Broad_LINCS_sig_info_2017-03-06.txt.gz",
            sep="\t",
            index_col=[0],
        )
        pert_dose = (
            sig_info["pert_idose"].apply(lambda x: x.replace(" um", "").strip()).astype(float)
        )
        sig_info["pert_idose_uM"] = np.where(pert_dose < 0, np.nan, pert_dose)
        gene_info = pd.read_csv(
            self.raw_data_dir / "GSE70138_Broad_LINCS_gene_info_2017-03-06.txt.gz",
            sep="\t",
            index_col=[0],
        )

        # Load in CMAP large dataset from GCTX file, reading it as an HDF5 file, and then walking the correct path.
        gz_f = gzip.open(self.raw_data_dir / CMAP_GCTX_GZ_FILENAME)
        f = h5py.File(gz_f)
        print(f"{datetime.datetime.now()} Extracting large matrix... ")
        big_df = pd.DataFrame(
            data=f["0"]["DATA"]["0"]["matrix"],
            columns=f["0"]["META"]["ROW"]["id"],
            index=f["0"]["META"]["COL"]["id"],
        )

        big_df.index = big_df.index.map(lambda x: x.decode("utf8"))
        big_df.columns = big_df.columns.map(lambda x: x.decode("utf8"))
        print(f"{datetime.datetime.now()} ..done")

        # Drop non-landmark features/genes if needed (by default we do)
        if self.landmark_only:
            big_df = big_df.drop(
                columns=[str(g) for g in gene_info.query("pr_is_lm==0").index.tolist()]
            ).rename(columns={str(n): gene_info.loc[n, "pr_gene_symbol"] for n in gene_info.index})

        # Define and write out features
        features = [gn for gn in big_df.columns if gn in gene_info.pr_gene_symbol.unique().tolist()]

        print("Adding metadata")
        old_columns = sig_info.columns.values.tolist()
        ssi = sig_info.distil_id.str.split(
            "|",
            expand=True,
        )
        sig_info = pd.concat([sig_info, ssi], axis=1)
        sig_info = sig_info.rename(
            columns={
                v: f"sample_id_{v}"
                for v in [cn for cn in sig_info.columns.values.tolist() if cn not in old_columns]
            }
        )
        new_siginfo = pd.concat(
            [
                sig_info.set_index(f"sample_id_{n}").drop(
                    columns=[
                        c
                        for c in sig_info.columns
                        if c.startswith("sample_id_") and c != f"sample_id_{n}"
                    ]
                    + ["distil_id"]
                )
                for n in range(6)
            ]
        ).drop(index=[None])
        big_df = big_df.join(new_siginfo)
        big_df = (
            big_df.query("pert_type in @self.allowed_treatment_types")
            .query("pert_itime in @self.allowed_treatment_times")
            .drop(columns=["pert_type", "pert_itime", "pert_idose"])
            .fillna({"pert_idose_uM": 0})
        )

        big_df["plate_id"] = big_df.index.map(lambda x: x.split(":")[0])
        big_df["well"] = big_df.index.map(lambda x: x.split(":")[-1])

        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=0)
        h5_store["df"] = big_df

        h5_store["gene_info"] = gene_info
        h5_store["sig_metrics"] = pd.read_csv(
            self.raw_data_dir / "GSE70138_Broad_LINCS_sig_metrics_2017-03-06.txt.gz",
            sep="\t",
            index_col=[0],
        )
        h5_store["pert_info"] = pd.read_csv(
            self.raw_data_dir / "GSE70138_Broad_LINCS_pert_info_2017-03-06.txt.gz",
            sep="\t",
            index_col=[0],
        )

        h5_store["inst_info"] = pd.read_csv(
            self.raw_data_dir / "GSE70138_Broad_LINCS_inst_info_2017-03-06.txt.gz",
            sep="\t",
            index_col=[0],
        )

        h5_store["features"] = pd.Series(features)
        h5_store["allowed_treatment_times"] = pd.Series(self.allowed_treatment_times)
        h5_store["allowed_treatment_types"] = pd.Series(self.allowed_treatment_types)

        h5_store["creation_date"] = pd.Series(str(datetime.datetime.now()))
        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        if key == "ds":
            key = "df"
        store = pd.HDFStore(self.processed_h5_file)
        pert_info_columns = [
            "pert_id",
            "pert_iname",
            "pert_type",
            "cell_id",
            "pert_idose",
            "pert_itime",
            "distil_id",
        ]
        df = store["/" + key]
        features = store["/features"].tolist()
        ds = Dataset("cmap", df, {"features": features})
        store.close()
        return ds



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from phenonaut.data import Dataset


def show_pair_plot(self, ds: Dataset):
    """Visualise the dataset using a pairplot.

    Parameters
    ----------
    ds : Dataset
        The dataset to visualise
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.pairplot(ds.df)
    plt.show()



# Copyright © The University of Edinburgh, 2023.
# Development has been supported by GSK.
import phenonaut
import pandas as pd
import phenonaut.transforms


def test_transforms_standardscaler(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    stdscaler = phenonaut.transforms.StandardScaler()
    stdscaler(phe.ds)
    assert (
        phe.df["StdScaler_feat_3"]
        - pd.Series([-0.505442, -0.424355, 0.994667, 1.724450, -0.910877, -0.878443])
    ).abs().mean() < 1e-6


def test_transforms_pca(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    t_pca = phenonaut.transforms.PCA()
    t_pca(phe.ds)
    assert (
        phe.df["PC_2"] - pd.Series([0.033607, -0.005505, 0.342870, -0.276157, -0.037338, -0.057477])
    ).abs().mean() < 1e-6


def test_transforms_zca(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    t_zca = phenonaut.transforms.ZCA()
    t_zca(phe.ds)
    assert (
        phe.df["ZCA_feat_3"]
        - pd.Series([-0.386266, -0.243655, -0.784378, 1.998697, -0.365108, -0.219291])
    ).abs().mean() < 1e-6


def test_transforms_tnse(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    constructor_kwargs = {}
    constructor_kwargs["perplexity"] = 2
    t_tsne = phenonaut.transforms.TSNE(constructor_kwargs, ndims=2)
    t_tsne(phe.ds)
    # Due to the stochastic nature of t-SNE, would not be a good idea to try to compare expected values
    assert len(phe.ds.features) == 2


def test_transforms_umap(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    t_umap = phenonaut.transforms.UMAP()
    t_umap(phe.ds)
    assert len(phe.ds.features) == 2


def test_transforms_custom_callable_square(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    from numpy import square, all as np_all

    f1_series = phe.ds.data.values[:, 0]
    t_square = phenonaut.transforms.Transformer(square)
    t_square(phe.ds)
    assert np_all(f1_series * f1_series == phe.ds.data.values[:, 0])


def test_transforms_robustmad(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    robust_mad = phenonaut.transforms.RobustMAD()

    # Enumerate all ways of calling robustmad:
    robust_mad(phe.ds)
    robust_mad.fit_transform(phe.ds)
    robust_mad.transform(phe.ds)
    robust_mad.fit(phe.ds)
    robust_mad(phe.ds, groupby="BARCODE")
    assert (
        phe.df["RobustMAD_RobustMAD_RobustMAD_RobustMAD_feat_3"]
        - pd.Series([-0.063631, 0.063631, 2.290720, 3.436080, -0.699942, -0.649037])
    ).abs().sum() < 1e-6


def test_transforms_log2(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    log2_t = phenonaut.transforms.Log2()
    log2_t(phe.ds)
    assert (
        phe.df["log2_feat_3"]
        - pd.Series([0.378512, 0.584963, 2.321928, 2.765535, -1.736966, -1.395929])
    ).abs().sum() < 2e-6



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from pandas import DataFrame
from phenonaut.data import Dataset
import numpy as np
from typing import Union, Optional

from phenonaut.phenonaut import Phenonaut


def _scalar_projection_and_rejection(a: np.ndarray, b: np.ndarray):
    """Return two scalar projection and rejection lists of a(s) onto b.

    Scalar projection and rejection calculated from:
    https://en.wikipedia.org/w/index.php?title=Scalar_projection&oldid=1038765889

    Projection of a onto unit vector of b (scalar projection) gets the length of
    components of a which are along b. Scalar rejection is therefore a-b,
    capturing components of a which are not along unit vector b and returning
    the length.  b is typically the target phenotype to which on off- values for
    phenotypes in a are assigned.

    a can be MxN numpy list, whereby M samples with N features have their scalar
    projections and rejections calculated, and returned as two arrays.
    Alternatively, a can be a 1D array capturing one sigle row of phenotypic
    features.

    Parameters
    ----------
    a : np.ndarray
        Query vector
    b : np.ndarray
        Target vector

    Returns
    -------
    tuple(np.ndarray, np.ndarray)
        Returns two numpy arrays, the first being a list of scalar projection
        values for M samples and the second being the scalar rejection values
        for the M samples.
    """
    # Check input is an array, as if a list, this can cause issues whereby [1,2]*2
    # produces a longer list, not multiplying elements, also we want to iterate
    # rows, so if 1d, change to 2D.
    input_shape = a.shape
    a = np.array(a)
    b = np.array(b)
    if len(input_shape) == 1:
        a = a.reshape(1, -1)
    scalar_projection_values = np.empty(a.shape[0])
    scalar_rejection_values = np.empty(a.shape[0])

    # We need a unit vector of B - strange that nympy does not have a single
    # function to do this.
    unit_vec = lambda x: x / np.linalg.norm(x)

    for i, row in enumerate(a):
        scalar_projection_values[i] = np.dot(row, unit_vec(b))

        # a1 (using the nomenclarute of the wikipedia page) ius components of A which
        # are not along unit vector b.
        a1 = np.dot(row, b) / (np.dot(b, b)) * b
        scalar_rejection_values[i] = np.linalg.norm(row - a1)

    if len(input_shape) == 1:
        return scalar_projection_values.item(0), scalar_rejection_values.item(0)
    return (scalar_projection_values, scalar_rejection_values)


def _scalar_projection(a: np.ndarray, b: np.ndarray):
    """Return scalar projection of a(s) onto b

    Calls _scalar_projection_and_rejection and takes the first returned tuple
    value.

    See _scalar_projection_and_rejection for further information

    Parameters
    ----------
    a : np.ndarray
        Query vector
    b : np.ndarray
        Target vector

    Returns
    -------
    Union[np.ndarray, float]
        Returns a numpy array or single float value giving the scalar
        projection
        values for M samples and the second being the scalar rejection values
        for the M samples.
    """
    return _scalar_projection_and_rejection(a, b)[0]


def _scalar_rejection(a: np.ndarray, b: np.ndarray):
    """Return scalar rejection of a(s) onto b.

    Calls _scalar_projection_and_rejection and takes the second returned tuple
    value.

    See _scalar_projection_and_rejection for further information

    Parameters
    ----------
    a : np.ndarray
        Query vector
    b : np.ndarray
        Target vector

    Returns
    -------
    Union[np.ndarray, float]
        Returns a numpy array or single float value giving the scalar
        projection
        values for M samples and the second being the scalar rejection values
        for the M samples.
    """
    return _scalar_projection_and_rejection(a, b)[1]


def _scalar_projection_scaled(a: np.ndarray, b: np.ndarray):
    """Return scalar projection of a(s) onto b, normalised to the length of b

    Calls _scalar_projection_and_rejection and takes the first returned tuple
    value.

    See _scalar_projection_and_rejection for further information

    Parameters
    ----------
    a : np.ndarray
        Query vector
    b : np.ndarray
        Target vector

    Returns
    -------
    Union[np.ndarray, float]
        Returns a numpy array or single float value giving the scalar
        projection
        values for M samples and the second being the scalar rejection values
        for the M samples.
    """
    return _scalar_projection_and_rejection(a, b)[0] / np.linalg.norm(b)


def _scalar_rejection_scaled(a: np.ndarray, b: np.ndarray):
    """Return scalar rejection of a(s) onto b, normalised to the length of b

    Calls _scalar_projection_and_rejection and takes the second returned tuple
    value.

    See _scalar_projection_and_rejection for further information

    Parameters
    ----------
    a : np.ndarray
        Query vector
    b : np.ndarray
        Target vector

    Returns
    -------
    Union[np.ndarray, float]
        Returns a numpy array or single float value giving the scalar
        projection
        values for M samples and the second being the scalar rejection values
        for the M samples.
    """
    return _scalar_projection_and_rejection(a, b)[1] / np.linalg.norm(b)


def scalar_projection(
    dataset: Dataset,
    target_perturbation_column_name="control",
    target_perturbation_column_value="pos",
    output_column_label="pos",
    norm=True,
):
    """Calculates the scalar projection and scalar rejection, quantifying on and
        off target phenotypes, as used in:
        Heiser, Katie, et al. "Identification of potential treatments for
        COVID-19 through artificial intelligence-enabled phenomic analysis of
        human cells infected with SARS-CoV-2." BioRxiv (2020).

    Parameters
    ----------
    dataset : Dataset
        Phenonaut dataset being queries
    target_TreatmentID : [type]
        TreatmentID of target phenotype. The median of all features across wells
        containing this phenotype is used.
    """

    on_phenotype_feature_label = f"on_target_{output_column_label}"
    off_phenotype_feature_label = f"off_target_{output_column_label}"

    # Target phenotype, against which all else is measured, calculated as the
    # median of all features from all wells carrying target phenotype perturbation.
    median_target_phenotype = dataset.df.loc[
        dataset.df.eval(
            f"{target_perturbation_column_name} == '{target_perturbation_column_value}'"
        ),
        dataset.features,
    ].median(axis=0)

    if median_target_phenotype.isnull().values.any():
        raise ValueError(f"Target phenotype contained null values: {median_target_phenotype}")

    # a is a point in phenotypic space - actually the entire dataframe as all is
    # calculated in one call.
    a = dataset.data
    # get scalar projection and rejection values.
    (
        scalar_projection_values,
        scalar_rejection_values,
    ) = _scalar_projection_and_rejection(a, median_target_phenotype)
    dataset.df[on_phenotype_feature_label] = scalar_projection_values / np.linalg.norm(
        median_target_phenotype
    )
    dataset.df[off_phenotype_feature_label] = scalar_rejection_values / np.linalg.norm(
        median_target_phenotype
    )

    dataset.features = (
        [on_phenotype_feature_label, off_phenotype_feature_label],
        f"Applied ScalarProjection in comparison with median phentype of {output_column_label}",
    )


# corrcoef_features_to_target
def feature_correlation_to_target(
    dataset: Union[Dataset, Phenonaut, DataFrame],
    target_feature: str,
    features: Optional[list[str]] = None,
    method: str = "pearson",
    return_dataframe: bool = True,
):
    """
    Calculate correlation coefficients for features to a column

    Sometimes we may wish to identify highly correlated features with a given
    property.

    In this example, we use a subset of the Iris dataset:

    ================= ================ ================= ================ ======
    sepal length (cm) sepal width (cm) petal length (cm) petal width (cm) target
    ================= ================ ================= ================ ======
    5.4                3.4               1.7              0.2              0
    7.2                3.0               5.8               1.6             2
    6.4                2.8               5.6               2.1             2
    4.8                3.1               1.60               2              0
    5.6                2.5               3.9               1.1             1
    ================= ================ ================= ================ ======

    We may wish to determine which feature is correlated with the
    "petal length (cm)" which can be achieved through calling this
    feature_correlation_to_target function, alowing the return of a pd.DataFrame
    or simple dictionary containing features names as keys, and the coefficients
    as values.

    .. code-block:: python

        import tempfile
        from phenonaut import Phenonaut
        with tempfile.NamedTemporaryFile(mode = "w") as tmp:
            tmp.write("sepal length (cm),sepal width (cm),petal length (cm),petal width (cm),target\\n5.4,3.4,1.7,0.2,0\\n7.2,3.0,5.8,1.6,2\\n6.4,2.8,5.6,2.1,2\\n4.8,3.1,1.60,2,0\\n5.6,2.5,3.9,1.1,1\\n")
            tmp.flush()
            phe=Phenonaut()
            phe.load_dataset("Flowers",tmp.name, {'features_regex':".*(width|length).*"})

        from phenonaut.metrics import feature_correlation_to_target
        print(feature_correlation_to_target(phe, 'petal length (cm)'))

    Returns a pd.DataFrame containing correlation coefficients:

    ================= ================================
    index             correlation_to_petal length (cm)
    ================= ================================
    sepal length (cm) 0.914639
    petal width (cm)  0.448289
    sepal width (cm)  -0.544665
    ================= ================================

    The optional dictionary, returned by calling the function with the additional return_dataframe parameter set to False:

    .. code-block:: python

        print(feature_correlation_to_target(phe, 'petal length (cm)'), return_dataframe=False)

    has the form:

    .. code-block:: python

        {'petal width (cm)': 0.448289248746271, 'sepal length (cm)': 0.9146393603234955, 'sepal width (cm)': -0.5446646166252519}

    Parameters
    ----------
    dataset : Union[Dataset, Phenonaut, DataFrame]
        The Phenonaut Dataset, pd.DataFrame, or Phenonaut object (containing only
        one dataset) on which to perform correlation calculations.
    target_feature : str
        The feature, or metadata column that all correlations should be calculated against.
    features : Optional[list[str]], optional
        List of features to include in the correlation calculations. If None,
        and a Dataset is supplied then those features are used. In the case where
        a pd.DataFrame is supplied, then features must be supplied. By default None.
    method : str, optional
        Method used to calculate the correlation coefficient. Can be 'pearson',
        'kendall' for the Kendall Tau correlation coefficient, or 'spearman' for
        the Spearman rank correlation. By default "pearson".
    return_dataframe : bool, optional
        If True, then a pd.DataFrame containing correlations is returned. If False,
        then a dictionary is returned, containing features names as keys, and
        the coefficients as values. By default True.

    Returns
    -------
    Union[pd.Datframe, dict]
        Return a pd.DataFrame with calculated correlation coefficients, or
        alternatively, a dictionary containing features names as keys, and the
        coefficients as values.

    Raises
    ------
    ValueError
        Phenonaut objects must contain only one dataset if passed to this
        function.
    ValueError
        Target feature not found in the supplied dataset.
    ValueError
        DataFrame supplied, but no features.
    ValueError
        target_feature not found in the dataset.
    TypeError
        Supplied dataset was not of type Phenonaut, Dataset, or pd.DataFrame
    """

    import pandas as pd

    # If Phenonaut object supplied, change it to the dataset
    if isinstance(dataset, Phenonaut):
        if len(dataset.datasets) != 1:
            raise ValueError(
                f"Phenonaut object passed to feature_correlation_to_target but it did not contain 1 dataset, it contained {len(dataset.datasets)}"
            )
        dataset = dataset[-1]

    # If Dataset, then make df with just required features
    if isinstance(dataset, Dataset):
        if features is None:
            features = list(dataset.features)
        if target_feature in features:
            features.remove(target_feature)
        if target_feature not in dataset.df.columns:
            raise ValueError("Given target_feature was not found in the supplied Dataset")
        dataset = dataset.df[features + [target_feature]]

    elif isinstance(dataset, DataFrame):
        if features is None:
            raise ValueError(
                "DataFrame provided to feature_correlation_to_target, but no features_list"
            )
        if target_feature not in dataset.columns:
            raise ValueError("Given target_feature was not found in the supplied Dataset")
        dataset = dataset[features + [target_feature]]
    else:
        raise TypeError(
            "dataset argument to feature_correlation_to_target was not of type Phenonaut, Dataset, or DataFrame"
        )

    coefs = {
        column_name: dataset[column_name].corr(dataset[target_feature], method=method)
        for column_name in features
    }

    if return_dataframe:
        return (
            DataFrame.from_dict(
                {"index": coefs.keys(), f"correlation_to_{target_feature}": coefs.values()}
            )
            .set_index("index")
            .sort_values(by=f"correlation_to_{target_feature}", ascending=False)
        )
    else:
        return coefs



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from sklearn.ensemble import GradientBoostingRegressor
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression
from sklearn.dummy import DummyRegressor
from phenonaut.predict.default_predictors.pytorch_models import MultiRegressorNN, DAVE
from phenonaut.predict.predictor_dataclasses import (
    PhenonautPredictor,
    HyperparameterInt,
    HyperparameterFloat,
    HyperparameterCategorical,
    HyperparameterLog,
)

default_multiregressors = [
    PhenonautPredictor("DummyRegressor-Average", (MultiOutputRegressor, DummyRegressor)),
    PhenonautPredictor(
        "MultiOutputRegressor-LinearRegression", (MultiOutputRegressor, LinearRegression)
    ),
    PhenonautPredictor(
        "MultiOutputRegressor-GradientBoosting",
        (MultiOutputRegressor, GradientBoostingRegressor),
        [
            HyperparameterFloat("learning_rate", 0.05, 0.2),
            HyperparameterInt("n_estimators", 1, 200),
        ],
    ),
    PhenonautPredictor(
        "MultiRegressorNN",
        MultiRegressorNN,
        [
            HyperparameterLog("learning_rate", 1e-5, 1e-1),
            HyperparameterInt("epochs", 1, 50),
            HyperparameterInt("num_hidden_layers", 1, 10),
            HyperparameterCategorical("use_optimizer", ["ADAM", "SGD"]),
        ],
        # conditional_hyperparameter_generator_constructor_keyword="hidden_layer_sizes",
        # conditional_hyperparameter_generator=lambda trial, optuna_dict, num_features_in, num_features_out: [
        #     trial.suggest_int(
        #         f"nodes_in_hidden_{n+1}",
        #         min(num_features_in, num_features_out),
        #         max(num_features_in, num_features_out) * 2,
        #     )
        #     for n in range(optuna_dict["num_hidden_layers"])
        # ],
    ),
    PhenonautPredictor(
        "DAVE",
        DAVE,
        [
            HyperparameterInt("num_hidden_layers", 1, 10),
            HyperparameterLog("learning_rate", 1e-5, 1e-1),
            HyperparameterCategorical("batch_size", [64, 128, 256, 512, 1024]),
            HyperparameterInt("epochs", 1, 20),
        ],
    ),
]



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

import phenonaut
from phenonaut.data.dataset import Dataset
import numpy as np
from phenonaut.phenonaut import Phenonaut


def test_get_features_and_non_features(dataset_iris):
    features = dataset_iris.features
    print(features)
    non_features = dataset_iris.get_non_feature_columns()
    assert sorted(features + non_features) == sorted(dataset_iris.df.columns.values)


def test_new_aggregated_dataset(small_2_plate_df):
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2

    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a sing value

    """
    phe = phenonaut.Phenonaut(small_2_plate_df)

    new_ds = phe.ds.new_aggregated_dataset(["ROW", "COLUMN", "BARCODE"])
    assert set(new_ds.features) == set(["feat_1", "feat_2", "feat_3"])
    assert abs(new_ds.df["feat_1"][0] - 1.25) < 1e-6
    assert np.abs(new_ds.df["feat_1"][1] - 5.70) < 1e-6
    assert np.abs(new_ds.df["feat_1"][2] - 0.15) < 1e-6
    assert np.abs(new_ds.df["feat_2"][0] - 1.3) < 1e-6
    assert np.abs(new_ds.df["feat_2"][1] - 5.6) < 1e-6
    assert np.abs(new_ds.df["feat_2"][2] - 0.2) < 1e-6
    assert np.abs(new_ds.df["feat_3"][0] - 1.4) < 1e-6
    assert np.abs(new_ds.df["feat_3"][1] - 5.9) < 1e-6
    assert np.abs(new_ds.df["feat_3"][2] - 0.34) < 1e-6


def test_iris_packageddataset():
    import tempfile

    tmpdir = tempfile.gettempdir()
    from phenonaut.packaged_datasets import Iris, Iris_2_views

    phe = Phenonaut(Iris(tmpdir))
    assert len(phe.keys()) == 1


def test_feature_selection_when_loading_csv():
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w") as tmp:
        tmp.write(
            "row, column, feature1, feature2, derived_feature3, feature10, feature003\n1,1,1,2,3,4,5\n2,1,6,7,8,9,10"
        )
        tmp.flush()
        phe = Phenonaut()
        phe.load_dataset(tmp.name, tmp.name, {"features_prefix": "feat"})
        assert phe.ds.features == ["feature1", "feature2", "feature003", "feature10"]

    with tempfile.NamedTemporaryFile(mode="w") as tmp:
        tmp.write(
            "sepal length (cm),sepal width (cm),petal length (cm),petal width (cm),target\n5.4,3.4,1.7,0.2,0\n7.2,3.0,5.8,1.6,2\n6.4,2.8,5.6,2.1,2\n4.8,3.1,1.60.2,0\n5.6,2.5,3.9,1.1,1"
        )
        tmp.flush()
        phe = Phenonaut()
        phe.load_dataset(tmp.name, tmp.name, {"features_regex": ".*(width|length).*"})
        assert phe.ds.features == [
            "petal length (cm)",
            "petal width (cm)",
            "sepal length (cm)",
            "sepal width (cm)",
        ]


def test_replace_string_in_column(small_2_plate_df):
    phe = phenonaut.Phenonaut(small_2_plate_df)
    phe[-1].replace_str("BARCODE", "Plate", "P")
    assert all(elem in ["P1", "P2"] for elem in phe.df.BARCODE.unique())


def test_remove_features_with_outliers(dataset_iris):
    dataset_iris.remove_features_with_outliers(7.5)
    assert len(dataset_iris.features) == 3


def test_remove_low_variance_features(dataset_iris):
    dataset_iris.remove_low_variance_features(freq_cutoff=0.5, unique_cutoff=0.2)
    assert dataset_iris.features == ["petal length (cm)", "sepal length (cm)"]


def test_remove_blocklist_features(dataset_iris):
    dataset_iris.remove_blocklist_features(["petal length (cm)", "sepal length (cm)"])
    assert dataset_iris.features == ["petal width (cm)", "sepal width (cm)"]
    dataset_iris.rename_column("petal width (cm)", "Nuclei_Correlation_Manders_AGP_DNA")
    dataset_iris.remove_blocklist_features("CellProfiler")
    assert dataset_iris.features == ["sepal width (cm)"]
